/*******************************************************************************

Copyright (c) 1997-2004, Perforce Software, Inc.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1.  Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

2.  Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTR
IBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL PERFORCE SOFTWARE, INC. BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/

/*******************************************************************************
 * Name		: php_clientapi.cc
 *
 * Author	: Tony Smith <tony@perforce.com> or <tony@smee.org>
 *
 * Description	: Php bindings for the Perforce API. Main interface to the
 * 		  Perforce API.
 *
 ******************************************************************************/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"
#include "php_ini.h"

BEGIN_EXTERN_C()
#include "ext/standard/info.h"
END_EXTERN_C()

#ifdef ZEND_ENGINE_2
#include "zend_object_handlers.h"
#endif

#include <clientapi.h>
#include <spec.h>
#include <diff.h>
#include <strtable.h>
#include "p4result.h"
#include "php_clientuser.h"
#include "php_clientapi.h"


/*******************************************************************************
 * Local functions
 ******************************************************************************/

static zval *	MkBool( int b )
{
    zval *z;
    MAKE_STD_ZVAL( z );
    if( b )
    {
	ZVAL_TRUE( z );
    }
    else
    {
	ZVAL_FALSE( z );
    }
    return z;
}

/*******************************************************************************
 * Our Php classes.
 ******************************************************************************/
#define		M_TAGGED		0x01
#define		M_PARSE_FORMS		0x02
#define		IS_TAGGED(x)		(x & M_TAGGED )
#define		IS_PARSE_FORMS(x)	(x & M_PARSE_FORMS )


PhpClientApi::PhpClientApi()
{
    initCount = 0;
    debug = 0;
    server2 = 0;
    mode = 0;
}

PhpClientApi::~PhpClientApi()
{
    if ( initCount )
    {
	Error e;
	client.Final( &e );
	// Ignore errors
    }
}

void
PhpClientApi::SetProg( const char *p )
{
#if P4APIVER_ID >= 513026
    // ClientApi::SetProg() was introduced in 2004.2
    client.SetProg( p );
#endif
}

void 
PhpClientApi::SetDebug( int d )
{
    debug = d;
    ui.SetDebug( d ? d - 1 : 0 );
}

//
// connect to the Perforce server.
//

int
PhpClientApi::Connect()
{
    if ( debug )
	fprintf( stderr, "[P4] Connecting to Perforce\n" );

    if ( initCount )
    {
	php_error(E_NOTICE, "P4.connect - Perforce client already connected!" );
	return 1;
    }

    Error	e;

    specDict.Clear();	// Don't cache specs across connections

    client.Init( &e );

    if ( e.Test() )
    {
	ReportError( &e );
	return 0;
    }

    initCount++;
    return 1;
}


//
// Disconnect session
//
int
PhpClientApi::Disconnect()
{
    if ( debug )
	fprintf( stderr, "[P4] Disconnect\n" );

    if ( ! initCount )
    {
	php_error( E_NOTICE, "P4#disconnect - not connected" );
	return 1;
    }

    Error	e;
    client.Final( &e );

    initCount--;
    return !e.Test();
}

//
// Test whether or not the connection has been dropped
//
int
PhpClientApi::Dropped()
{
    return client.Dropped();
}

void 
PhpClientApi::Tagged()
{
    client.SetProtocol( "tag", "" );
    mode |= M_TAGGED;
}

void 
PhpClientApi::ParseForms()
{
    client.SetProtocol( "tag", "" );
    client.SetProtocol( "specstring", "" );
    mode |= M_PARSE_FORMS + M_TAGGED ;
}

//
// Run returns the results of the command as an array. Errors are reported
// through the Errors() interface.
//

zval *
PhpClientApi::Run( const char *cmd, int argc, char * const *argv )
{
    // Save the entire command string for our error messages. Makes it
    // easy to see where a script has gone wrong.
    StrBuf	cmdString;
    cmdString << "\"p4 " << cmd;
    for( int i = 0; i < argc; i++ )
        cmdString << " " << argv[ i ];
    cmdString << "\"";

    // Clear out any results from the previous command
    ui.Reset();

    if ( debug )
	fprintf( stderr, "[P4] Executing %s\n", cmdString.Text()  );

    if ( !initCount )
    {
	ui.AddError( "P4.run: not connected." );
	return GetOutput();
    }

    RunCmd( cmd, &ui, argc, argv );

    if( ui.LastSpecDef().Length() )
	specDict.SetVar( cmd, ui.LastSpecDef() );

    return GetOutput();
}


//
// RunCmd is a private function to work around an obscure protocol
// bug in 2000.[12] servers. Running a "p4 -Ztag client -o" messes up the
// protocol so if they're running this command then we disconnect and
// reconnect to refresh it. For efficiency, we only do this if the 
// server2 protocol is either 9 or 10 as other versions aren't affected.
//
void
PhpClientApi::RunCmd( const char *cmd, ClientUser *ui, int argc, char * const *argv )
{
    client.SetArgv( argc, argv );
    client.Run( cmd, ui );

    // Have to request server2 protocol *after* a command has been run. I
    // don't know why, but that's the way it is.

    if ( ! server2 )
    {
	StrPtr *pv = client.GetProtocol( "server2" );
	if ( pv )
	    server2 = pv->Atoi();
    }

    if ( IS_TAGGED(mode) && StrRef( cmd ) == "client" && 
	 server2 >= 9    && server2 <= 10  )
    {
	if ( argc && ( StrRef( argv[ 0 ] ) == "-o" ) )
	{
	    if ( debug )
		printf( "Resetting client to avoid 2000.[12] protocol bug\n" );

	    Error e;
	    client.Final( &e );
	    client.Init( &e );

	    // Pass any errors down to the UI, so they'll get picked up.
	    if ( e.Test() ) 
		ui->HandleError( &e );
	}
    }
}


//
// Parses a string supplied by the user into a hash. To do this we need
// the specstring from the server. We try to cache those as we see them, 
// but the user may not have executed any commands to allow us to cache
// them so we may have to fetch the spec first.
//

zval *
PhpClientApi::ParseSpec( const char * type, const char *form )
{
    StrPtr	*specDef = FetchSpecDef( type );

    if ( !specDef )
    {
	StrBuf m;
	m = "No spec definition for ";
	m.Append( type );
	m.Append( " objects." );
	ui.AddError( m.Text() );
	return MkBool( 0 );
    }

    // Got a specdef so now we can attempt to parse it.
    
    SpecDataTable	specData;
    Spec		s( specDef->Text(), "" );
    Error		e;

    s.ParseNoValid( form, &specData, &e );
    if ( e.Test() )
    {
	ui.HandleError( &e );
	return MkBool( 0 );
    }

    // Now we've parsed it, convert it into a hash. We do that via 
    // direct access to the method in PhpClientUser - this is ugly, but
    // expedient.

    return ui.DictToHash( specData.Dict(), specDef );
}
    

//
// Supply the input for the next command
//
void
PhpClientApi::SetInput( zval *input )
{
    if ( debug )
	fprintf( stderr, "[P4] Received input for next command\n" );

    ui.SetInput( input );
}

//
// Fetch a spec definition from the cache - faulting it if it's not there.
//

StrPtr * 
PhpClientApi::FetchSpecDef( const char *type )
{
    StrPtr *sd = specDict.GetVar( type );
    if( sd ) return sd;

    // Fault. Now we have to do something nasty. If we're in parse_forms mode, 
    // we can run a "p4 XXXX -o" and discard the result - the specdef should 
    // now be in the cache. If we're not in parse_forms mode, we're hosed. 
    // There's just no other way to get the specdef.

    if( ! IS_PARSE_FORMS(mode) )
    {
	ReportError( "P4.parse_spec: Use parse_forms mode to parse specs." );
	return NULL;
    }

    char * const argv[] = { "-o" };
    Run( type, 1, argv );

    sd = specDict.GetVar( type );
    if( sd ) return sd;

    // OK, now we're hosed. 
    ReportError( "P4.parse_spec: Error getting spec definition from server!");
    return NULL;
}

//
// Report a Perforce error using php_error()
//
void
PhpClientApi::ReportError( Error *e )
{
    StrBuf	b;
    e->Fmt( &b );
    php_error( E_NOTICE, "Perforce Error: %s\n", b.Text() );
}

//
// Report an error using php_error()
//
void
PhpClientApi::ReportError( const char *msg )
{
    php_error( E_NOTICE, "Error: %s\n", msg );
}


